@@ -1,5 +1,6 @@
 import multiprocessing
 from matplotlib import pyplot as plt
+from numba import njit
 import numpy as np
 import pandas as pd
 import random
@@ -20,7 +21,9 @@ class LPPLS(object):
         self.coef_ = {}
         self.indicator_result = []
 
-    def lppls(self, t, tc, m, w, a, b, c1, c2):
+    @staticmethod
+    @njit
+    def lppls(t, tc, m, w, a, b, c1, c2):
         return a + np.power(tc - t, m) * (b + ((c1 * np.cos(w * np.log(tc - t))) + (c2 * np.sin(w * np.log(tc - t)))))
 
     def func_restricted(self, x, *args):
@@ -47,7 +50,9 @@ class LPPLS(object):
 
         return np.sum(delta)
 
-    def matrix_equation(self, observations, tc, m, w):
+    @staticmethod
+    @njit
+    def matrix_equation(observations, tc, m, w):
         """
         Derive linear parameters in LPPLs from nonlinear ones.
         """
@@ -58,9 +63,9 @@ class LPPLS(object):
         fi = np.power(deltaT, m)
         gi = fi * np.cos(w * phase)
         hi = fi * np.sin(w * phase)
-        A = np.stack([np.ones_like(deltaT), fi, gi, hi])
+        A = np.stack((np.ones_like(deltaT), fi, gi, hi))
 
-        return np.linalg.lstsq(A.T, P, rcond=None)[0].astype('float').tolist()
+        return np.linalg.lstsq(A.T, P, rcond=-1.0)[0]
 
     def fit(self, observations, max_searches, minimizer='Nelder-Mead'):
         """
@@ -77,11 +82,12 @@ class LPPLS(object):
         while search_count < max_searches:
             tc_init_min, tc_init_max = self._get_tc_bounds(observations, 0.20, 0.20)
 
+            # @TODO make configurable
             # set random initialization limits for non-linear params
             init_limits = [
                 (tc_init_min, tc_init_max),  # tc : Critical Time
-                (0.1, 0.9),  # m : 0.1 ≤ m ≤ 0.9
-                (6, 13),  # ω : 6 ≤ ω ≤ 13
+                (0, 2),  # m : 0.1 ≤ m ≤ 0.9
+                (1, 50),  # ω : 6 ≤ ω ≤ 13
             ]
 
             # randomly choose vals within bounds for non-linear params
@@ -162,32 +168,19 @@ class LPPLS(object):
         Returns:
             nothing, should plot the indicator
         """
-        price = self.observations[1, :]
-        n = len(price) - len(res)
-        pos_conf_lst = [0] * n
-        neg_conf_lst = [0] * n
-        for r in res:
-            pos_true_count = 0
-            neg_true_count = 0
-            for fits in r:
-                if fits['qualified'][condition_name] and fits['sign'] > 0:
-                    pos_true_count = pos_true_count + 1
-                if fits['qualified'][condition_name] and fits['sign'] < 0:
-                    neg_true_count = neg_true_count + 1
-            pos_conf_lst.append(pos_true_count / len(r))
-            neg_conf_lst.append(neg_true_count / len(r))
+        res_df = self.res_to_df(res, condition_name)
 
         fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, sharex=True, figsize=(15, 12))
         fig.suptitle(title)
         # plot pos bubbles
         ax1_0 = ax1.twinx()
-        ax1.plot(price, color='black')
-        ax1_0.plot(pos_conf_lst, label='bubble indicator (pos)')
+        ax1.plot(res_df['price'].values, color='black')
+        ax1_0.plot(res_df['pos_conf'].values, label='bubble indicator (pos)')
 
         # plot neg bubbles
         ax2_0 = ax2.twinx()
-        ax2.plot(price, color='black')
-        ax2_0.plot(neg_conf_lst, label='bubble indicator (neg)')
+        ax2.plot(res_df['price'].values, color='black')
+        ax2_0.plot(res_df['neg_conf'].values, label='bubble indicator (neg)')
 
         # set grids
         ax1.grid(which='major', axis='both', linestyle='--')
@@ -209,17 +202,19 @@ class LPPLS(object):
         obs_copy_len = len(obs_copy[0, :]) - window_size
 
         func = self._func_compute_indicator
+
         func_arg_map = [(
             obs_copy[:, i:window_size + i],  # obs
             i,  # n_iter
             window_size,  # window_size
             smallest_window_size,  # smallest_window_size
             increment,  # increment
-            max_searches, # max_searches
+            max_searches,  # max_searches
             filter_conditions_config,
         ) for i in range(obs_copy_len)]
 
         pool = multiprocessing.Pool(processes=workers)
+
         result = pool.map(func, func_arg_map)
         pool.close()
 
@@ -239,7 +234,7 @@ class LPPLS(object):
             obs_shrinking_slice = obs[:, j * increment:window_size + n_iter]
 
             # fit the model to the data and get back the params
-            tc, m, w, a, b, c, c1, c2 = self.fit(obs_shrinking_slice, max_searches, minimizer='Nelder-Mead')
+            tc, m, w, a, b, c, c1, c2 = self.fit(obs_shrinking_slice, max_searches, minimizer='SLSQP')
 
             first = obs_shrinking_slice[0][0]
             last = obs_shrinking_slice[0][-1]
@@ -311,3 +306,49 @@ class LPPLS(object):
         tc_init_min = t_last - pct_delta_min
         tc_init_max = t_last + pct_delta_max
         return tc_init_min, tc_init_max
+
+    def res_to_df(self, res, condition_name):
+        """
+        Args:
+            res (list): result from mp_compute_indicator
+            condition_name (str): the name you assigned to the filter condition in your config
+        Returns:
+            pd.DataFrame()
+        """
+        idx = self.observations[0, :]
+        price = self.observations[1, :]
+        n = len(price) - len(res)
+        # pos_fits = [0] * n
+        # neg_fits = [0] * n
+        pos_conf_lst = [0] * n
+        neg_conf_lst = [0] * n
+        fits_ = [0] * n
+
+        for r in res:
+            pos_count = 0
+            neg_count = 0
+            pos_true_count = 0
+            neg_true_count = 0
+            for fits in r:
+
+                if fits['sign'] > 0:
+                    pos_count += 1
+                    if fits['qualified'][condition_name]:
+                        pos_true_count += 1
+                if fits['sign'] < 0:
+                    neg_count += 1
+                    if fits['qualified'][condition_name]:
+                        neg_true_count = neg_true_count + 1
+            # pos_conf_lst.append(pos_true_count / len(r))
+            # neg_conf_lst.append(neg_true_count / len(r))
+            fits_.append(fits)
+            pos_conf_lst.append(pos_true_count / pos_count if pos_count > 0 else 0)
+            neg_conf_lst.append(neg_true_count / neg_count if neg_count > 0 else 0)
+
+        return pd.DataFrame({
+            'idx': idx,
+            'price': price,
+            'pos_conf': pos_conf_lst,
+            'neg_conf': neg_conf_lst,
+            'fit_params': fits_,
+        }).set_index('idx')
